@@ -12,29 +12,49 @@ bindings after the fact.
 We can also move the implementation to C++ if desired.
 """
 
+import datetime
 import inspect
+import mimetypes
+import platform
 import shutil
-from collections.abc import KeysView
+from collections.abc import KeysView, MutableMapping
 from decimal import Decimal
 from io import BytesIO
-from os import replace
 from pathlib import Path
 from subprocess import PIPE, run
 from tempfile import NamedTemporaryFile
-from typing import Any, BinaryIO, Callable, List, Optional, Tuple, Type, TypeVar, Union
+from typing import (
+    Any,
+    BinaryIO,
+    Callable,
+    ItemsView,
+    Iterator,
+    List,
+    Optional,
+    Tuple,
+    Type,
+    TypeVar,
+    Union,
+    ValuesView,
+)
 from warnings import warn
 
 from . import Array, Dictionary, Name, Object, Page, Pdf, Stream
 from ._qpdf import (
     AccessMode,
+    AttachedFile,
+    AttachedFileSpec,
+    Attachments,
+    NameTree,
     ObjectStreamMode,
-    PdfError,
+    Rectangle,
     StreamDecodeLevel,
     StreamParser,
     Token,
     _ObjectMapping,
 )
 from .models import Encryption, EncryptionInfo, Outline, PdfMetadata, Permissions
+from .models.metadata import decode_pdf_date, encode_pdf_date
 
 # pylint: disable=no-member,unsupported-membership-test,unsubscriptable-object
 # mypy: ignore-errors
@@ -44,6 +64,30 @@ __all__ = []
 Numeric = TypeVar('Numeric', int, float, Decimal)
 
 
+def augment_override_cpp(fn):
+    """This Python function should replace the C++ implementation, if there is one."""
+    fn._augment_override_cpp = True
+    return fn
+
+
+def augment_if_no_cpp(fn):
+    """This Python function provides a Python implementation if no C++ implementation exists."""
+    fn._augment_if_no_cpp = True
+    return fn
+
+
+def _is_inherited_method(meth):
+    # Augmenting a C++ with a method that cls inherits from the Python
+    # object is never what we want.
+    return meth.__qualname__.startswith('object.')
+
+
+def _is_augmentable(m):
+    return (
+        inspect.isfunction(m) and not _is_inherited_method(m)
+    ) or inspect.isdatadescriptor(m)
+
+
 def augments(cls_cpp: Type[Any]):
     """Attach methods of a Python support class to an existing class
 
@@ -72,7 +116,7 @@ def augments(cls_cpp: Type[Any]):
     Any existing methods may be used, regardless of whether they are defined
     elsewhere in the support class or in the target class.
 
-    For data fields to work, including @property accessors, the target class must be
+    For data fields to work, the target class must be
     tagged ``py::dynamic_attr`` in pybind11.
 
     Strictly, the target class does not have to be C++ or derived from pybind11.
@@ -82,20 +126,52 @@ def augments(cls_cpp: Type[Any]):
 
     (Alternative ideas: https://github.com/pybind/pybind11/issues/1074)
     """
-    ATTR_WHITELIST = {'__repr__', '__enter__', '__exit__'}
+
+    OVERRIDE_WHITELIST = {'__eq__', '__hash__', '__repr__'}
+    if platform.python_implementation() == 'PyPy':
+        # Either PyPy or pybind11's interface to PyPy automatically adds a __getattr__
+        OVERRIDE_WHITELIST |= {'__getattr__'}  # pragma: no cover
 
     def class_augment(cls, cls_cpp=cls_cpp):
-        for name, member in inspect.getmembers(cls):
-            # Don't replace existing methods except those in our whitelist
-            if hasattr(cls_cpp, name) and name not in ATTR_WHITELIST:
+
+        # inspect.getmembers has different behavior on PyPy - in particular it seems
+        # that a typical PyPy class like cls will have more methods that it considers
+        # methods than CPython does. Our predicate should take care of this.
+        for name, member in inspect.getmembers(cls, predicate=_is_augmentable):
+            if name == '__weakref__':
                 continue
+            if (
+                hasattr(cls_cpp, name)
+                and hasattr(cls, name)
+                and name not in getattr(cls, '__abstractmethods__', set())
+                and name not in OVERRIDE_WHITELIST
+                and not getattr(getattr(cls, name), '_augment_override_cpp', False)
+            ):
+                if getattr(getattr(cls, name), '_augment_if_no_cpp', False):
+                    # If tagged as "augment if no C++", we are supported older pybind11
+                    # versions that may not define some functions. If pybind11 provides
+                    # the required function, don't override it.
+                    continue
+
+                # If the original C++ class and Python support class both define the
+                # same name, we generally have a conflict, because this is augmentation
+                # not inheritance. However, if the method provided by the support class
+                # is an abstract method, then we can consider the C++ version the
+                # implementation. Also, pybind11 provides defaults for __eq__,
+                # __hash__ and __repr__ that we often do want to override directly.
+
+                raise RuntimeError(  # pragma: no cover
+                    f"C++ {cls_cpp} and Python {cls} both define the same "
+                    f"non-abstract method {name}: "
+                    f"{getattr(cls_cpp, name, '')!r}, "
+                    f"{getattr(cls, name, '')!r}"
+                )
             if inspect.isfunction(member):
-                if member.__qualname__.startswith('object.'):
-                    continue  # To avoid breaking PyPy
-                member.__qualname__ = member.__qualname__.replace(
+                setattr(cls_cpp, name, member)
+                installed_member = getattr(cls_cpp, name)
+                installed_member.__qualname__ = member.__qualname__.replace(
                     cls.__name__, cls_cpp.__name__
                 )
-                setattr(cls_cpp, name, member)
             elif inspect.isdatadescriptor(member):
                 setattr(cls_cpp, name, member)
 
@@ -110,7 +186,7 @@ def augments(cls_cpp: Type[Any]):
 
 
 def _single_page_pdf(page) -> bytes:
-    """Construct a single page PDF from the provided page in memory"""
+    """Construct a single page PDF from the provided page in memory."""
     pdf = Pdf.new()
     pdf.pages.append(page)
     bio = BytesIO()
@@ -120,7 +196,7 @@ def _single_page_pdf(page) -> bytes:
 
 
 def _mudraw(buffer, fmt) -> bytes:
-    """Use mupdf draw to rasterize the PDF in the memory buffer"""
+    """Use mupdf draw to rasterize the PDF in the memory buffer."""
     # mudraw cannot read from stdin so NamedTemporaryFile is required
     with NamedTemporaryFile(suffix='.pdf') as tmp_in:
         tmp_in.write(buffer)
@@ -138,20 +214,6 @@ def _mudraw(buffer, fmt) -> bytes:
 
 @augments(Object)
 class Extend_Object:
-    def _repr_mimebundle_(self, include=None, exclude=None):
-        """Present options to IPython for rich display of this object
-
-        See https://ipython.readthedocs.io/en/stable/config/integrating.html#rich-display
-        """
-
-        if (
-            isinstance(self, Dictionary)
-            and Name.Type in self
-            and self.Type == Name.Page
-        ):
-            return Page(self)._repr_mimebundle_(include, exclude)
-        return None
-
     def _ipython_key_completions_(self):
         if isinstance(self, (Dictionary, Stream)):
             return self.keys()
@@ -192,7 +254,7 @@ class Extend_Object:
             raise TypeError("Objects must have the same owner for emplace()")
 
         # .keys() returns strings, so make all strings
-        retain = set(str(k) for k in retain)
+        retain = {str(k) for k in retain}
         self_keys = set(self.keys())
         other_keys = set(other.keys())
 
@@ -204,6 +266,42 @@ class Extend_Object:
         for k in del_keys:
             del self[k]  # pylint: disable=unsupported-delete-operation
 
+    def _type_check_write(self, filter, decode_parms):
+        if isinstance(filter, list):
+            filter = Array(filter)
+        filter = filter.wrap_in_array()
+
+        if isinstance(decode_parms, list):
+            decode_parms = Array(decode_parms)
+        elif decode_parms is None:
+            decode_parms = Array([])
+        else:
+            decode_parms = decode_parms.wrap_in_array()
+
+        if not all(isinstance(item, Name) for item in filter):
+            raise TypeError(
+                "filter must be: pikepdf.Name or pikepdf.Array([pikepdf.Name])"
+            )
+        if not all(
+            (isinstance(item, Dictionary) or item is None) for item in decode_parms
+        ):
+            raise TypeError(
+                "decode_parms must be: pikepdf.Dictionary or "
+                "pikepdf.Array([pikepdf.Dictionary])"
+            )
+        if len(decode_parms) != 0 and len(filter) != len(decode_parms):
+            raise ValueError(
+                f"filter ({repr(filter)}) and decode_parms "
+                f"({repr(decode_parms)}) must be arrays of same length"
+            )
+        if len(filter) == 1:
+            filter = filter[0]
+        if len(decode_parms) == 0:
+            decode_parms = None
+        elif len(decode_parms) == 1:
+            decode_parms = decode_parms[0]
+        return filter, decode_parms
+
     def write(
         self,
         data: bytes,
@@ -246,57 +344,16 @@ class Extend_Object:
         """
 
         if type_check and filter is not None:
-            if isinstance(filter, list):
-                filter = Array(filter)
-            filter = filter.wrap_in_array()
+            filter, decode_parms = self._type_check_write(filter, decode_parms)
 
-            if isinstance(decode_parms, list):
-                decode_parms = Array(decode_parms)
-            elif decode_parms is None:
-                decode_parms = Array([])
-            else:
-                decode_parms = decode_parms.wrap_in_array()
-
-            if not all(isinstance(item, Name) for item in filter):
-                raise TypeError(
-                    "filter must be: pikepdf.Name or pikepdf.Array([pikepdf.Name])"
-                )
-            if not all(
-                (isinstance(item, Dictionary) or item is None) for item in decode_parms
-            ):
-                raise TypeError(
-                    "decode_parms must be: pikepdf.Dictionary or "
-                    "pikepdf.Array([pikepdf.Dictionary])"
-                )
-            if len(decode_parms) != 0:
-                if len(filter) != len(decode_parms):
-                    raise ValueError(
-                        f"filter ({repr(filter)}) and decode_parms "
-                        f"({repr(decode_parms)}) must be arrays of same length"
-                    )
-            if len(filter) == 1:
-                filter = filter[0]
-            if len(decode_parms) == 0:
-                decode_parms = None
-            elif len(decode_parms) == 1:
-                decode_parms = decode_parms[0]
         self._write(data, filter=filter, decode_parms=decode_parms)
 
 
 @augments(Pdf)
 class Extend_Pdf:
-    @property
-    def root(self):
-        """
-        Deprecated alias for .Root, the /Root object of the PDF.
-
-        .. deprecated:: 2.0
-            Use ``.Root``.
-        """
-        warn("Pdf.root is deprecated; use Pdf.Root", category=DeprecationWarning)
-        return self.Root
-
-    def _repr_mimebundle_(self, include=None, exclude=None):
+    def _repr_mimebundle_(
+        self, include=None, exclude=None
+    ):  # pylint: disable=unused-argument
         """
         Present options to IPython or Jupyter for rich display of this object
 
@@ -451,7 +508,7 @@ class Extend_Pdf:
 
     def add_blank_page(
         self, *, page_size: Tuple[Numeric, Numeric] = (612.0, 792.0)
-    ) -> Object:
+    ) -> Page:
         """
         Add a blank page to this PDF. If pages already exist, the page will be added to
         the end. Pages may be reordered using ``Pdf.pages``.
@@ -473,9 +530,9 @@ class Extend_Pdf:
             Contents=self.make_stream(b''),
             Resources=Dictionary(),
         )
-        page = self.make_indirect(page_dict)
-        self._add_page(page, first=False)
-        return page
+        page_obj = self.make_indirect(page_dict)
+        self._add_page(page_obj, first=False)
+        return Page(page_obj)
 
     def close(self) -> None:
         """
@@ -494,36 +551,11 @@ class Extend_Pdf:
         the destination ``Pdf`` immediately, so after accessing all desired information
         from the source ``Pdf`` it may be closed.
 
-        Caution:
-            Closing the ``Pdf`` is currently implemented by resetting it to an empty
-            sentinel. It is currently possible to edit the sentinel as if it were a live
-            object. This behavior should not be relied on and is subject to change.
+        .. versionchanged:: 3.0
+            In pikepdf 2.x, this function actually worked by resetting to a very short
+            empty PDF. Code that relied on this quirk may not function correctly.
         """
-
-        # We could use QPDF::closeInputSource(), but many functions like
-        # QPDF::getFilename() will segfault if called without an open file, so it's
-        # best to use a sentinel empty file.
-        EMPTY_PDF = (
-            b"%PDF-1.3\n"
-            b"1 0 obj\n"
-            b"<< /Type /Catalog /Pages 2 0 R >>\n"
-            b"endobj\n"
-            b"2 0 obj\n"
-            b"<< /Type /Pages /Kids [] /Count 0 >>\n"
-            b"endobj\n"
-            b"xref\n"
-            b"0 3\n"
-            b"0000000000 65535 f \n"
-            b"0000000009 00000 n \n"
-            b"0000000058 00000 n \n"
-            b"trailer << /Size 3 /Root 1 0 R >>\n"
-            b"startxref\n"
-            b"110\n"
-            b"%%EOF\n"
-        )
-
-        description = "closed file: " + self.filename
-        self._process(description, EMPTY_PDF)
+        self._close()
         if getattr(self, '_tmp_stream', None):
             self._tmp_stream.close()
 
@@ -568,7 +600,7 @@ class Extend_Pdf:
             def __init__(self):  # pylint: disable=useless-super-delegation
                 super().__init__()  # required for C++
 
-            def handle_object(self, obj):
+            def handle_object(self, *_args):
                 pass
 
             def handle_eof(self):
@@ -579,8 +611,7 @@ class Extend_Pdf:
         self._decode_all_streams_and_discard()
 
         discarding_parser = DiscardingParser()
-        for basic_page in self.pages:
-            page = Page(basic_page)
+        for page in self.pages:
             page.parse_contents(discarding_parser)
 
         for warning in self.get_warnings():
@@ -588,85 +619,10 @@ class Extend_Pdf:
 
         return problems
 
-    def _attach(
-        self,
-        *,
-        basename: str,
-        filebytes: bytes,
-        mime: Optional[str] = None,
-        desc: str = '',
-    ):  # pragma: no cover
-        """
-        Attach a file to this PDF
-
-        Args:
-            basename (str): The basename (filename withouth path) to name the
-                file. Not necessarily the name of the file on disk. Will be s
-                hown to the user by the PDF viewer. filebytes (bytes): The file
-                contents.
-
-            mime (str or None): A MIME type for the filebytes. If omitted, we try
-                to guess based on the standard library's
-                :func:`mimetypes.guess_type`. If this cannot be determined, the
-                generic value `application/octet-stream` is used. This value is
-                used by PDF viewers to decide how to present the information to
-                the user.
-
-            desc (str): A extended description of the file contents. PDF viewers
-                also display this information to the user. In Acrobat DC this is
-                hidden in a context menu.
-
-        The PDF will also be modified to request the PDF viewer to display the
-        list of attachments when opened, as opposed to other viewing modes. Some
-        PDF viewers will not make it obvious to the user that attachments are
-        present unless this is done. This behavior may be overridden by changing
-        ``pdf.Root.PageMode`` to some other valid value.
-
-        """
-
-        if '/Names' not in self.Root:
-            self.Root.Names = self.make_indirect(Dictionary())
-        if '/EmbeddedFiles' not in self.Root:
-            self.Root.Names.EmbeddedFiles = self.make_indirect(Dictionary())
-        if '/Names' not in self.Root.Names.EmbeddedFiles:
-            self.Root.Names.EmbeddedFiles.Names = Array()
-
-        if '/' in basename or '\\' in basename:
-            raise ValueError("basename should be a basename (no / or \\)")
-
-        if not mime:
-            from mimetypes import guess_type
-
-            mime, _encoding = guess_type(basename)
-            if not mime:
-                mime = 'application/octet-stream'
-
-        filestream = Stream(self, filebytes)
-        filestream.Subtype = Name('/' + mime)
-
-        filespec = Dictionary(
-            {
-                '/Type': Name.Filespec,
-                '/F': basename,
-                '/UF': basename,
-                '/Desc': desc,
-                '/EF': Dictionary({'/F': filestream}),
-            }
-        )
-
-        # names = self.Root.Names.EmbeddedFiles.Names.as_list()
-        # names.append(filename)  # Key
-        # names.append(self.make_indirect(filespec))
-        self.Root.Names.EmbeddedFiles.Names = Array(
-            [basename, self.make_indirect(filespec)]  # key
-        )
-
-        if '/PageMode' not in self.Root:
-            self.Root.PageMode = Name.UseAttachments
-
     def save(
-        self,  # TODO mandatory kwargs
+        self,
         filename_or_stream: Union[Path, str, BinaryIO, None] = None,
+        *,
         static_id: bool = False,
         preserve_pdfa: bool = True,
         min_version: Union[str, Tuple[str, int]] = "",
@@ -738,7 +694,7 @@ class Extend_Pdf:
 
             stream_decode_level: Specifies how
                 to encode stream objects. See documentation for
-                ``StreamDecodeLevel``.
+                :class:`pikepdf.StreamDecodeLevel`.
 
             recompress_flate: When disabled (the default), qpdf does not
                 uncompress and recompress streams compressed with the Flate
@@ -775,6 +731,7 @@ class Extend_Pdf:
         Raises:
             PdfError
             ForeignObjectError
+            ValueError
 
         You may call ``.save()`` multiple times with different parameters
         to generate different versions of a file, and you *may* continue
@@ -797,9 +754,22 @@ class Extend_Pdf:
 
         .. versionchanged:: 2.7
             Added *recompress_flate*.
+
+        .. versionchanged:: 3.0
+            Keyword arguments now mandatory for everything except the first
+            argument.
         """
-        if not filename_or_stream and self._original_filename:
+        if not filename_or_stream and getattr(self, '_original_filename', None):
             filename_or_stream = self._original_filename
+        if not filename_or_stream:
+            raise ValueError(
+                "Cannot save to original filename because the original file was "
+                "not opening using Pdf.open(..., allow_overwriting_input=True). "
+                "Either specify a new destination filename/file stream or open "
+                "with allow_overwriting_input=True. If this Pdf was created using "
+                "Pdf.new(), you must specify a destination object since there is "
+                "no original filename to save to."
+            )
         self._save(
             filename_or_stream,
             static_id=static_id,
@@ -820,8 +790,9 @@ class Extend_Pdf:
         )
 
     @staticmethod
-    def open(  # TODO mandatory kwargs
+    def open(
         filename_or_stream: Union[Path, str, BinaryIO],
+        *,
         password: Union[str, bytes] = "",
         hex_password: bool = False,
         ignore_xref_streams: bool = False,
@@ -920,6 +891,10 @@ class Extend_Pdf:
             perform poorly. It may be easier to download a PDF from network to
             temporary local storage (such as ``io.BytesIO``), manipulate it, and
             then re-upload it.
+
+        .. versionchanged:: 3.0
+            Keyword arguments now mandatory for everything except the first
+            argument.
         """
         if isinstance(filename_or_stream, bytes) and filename_or_stream.startswith(
             b'%PDF-'
@@ -940,7 +915,7 @@ class Extend_Pdf:
                     '"allow_overwriting_input=True" requires "open" first argument '
                     'to be a file path'
                 ) from error
-            original_filename = str(filename_or_stream)
+            original_filename = Path(filename_or_stream)
             with open(original_filename, 'rb') as pdf_file:
                 tmp_stream = BytesIO()
                 shutil.copyfileobj(pdf_file, tmp_stream)
@@ -954,8 +929,8 @@ class Extend_Pdf:
             inherit_page_attributes=inherit_page_attributes,
             access_mode=access_mode,
         )
-        setattr(pdf, '_tmp_stream', tmp_stream)
-        setattr(pdf, '_original_filename', original_filename)
+        pdf._tmp_stream = tmp_stream
+        pdf._original_filename = original_filename
         return pdf
 
 
@@ -967,26 +942,28 @@ class Extend_ObjectMapping:
         except KeyError:
             return default
 
+    @augment_if_no_cpp
     def keys(self):
         return KeysView(self)
 
+    @augment_if_no_cpp
     def values(self):
         return (v for _k, v in self.items())
 
 
-def check_is_box(obj) -> bool:
+def check_is_box(obj) -> None:
     try:
         if obj.is_rectangle:
-            return True
+            return
     except AttributeError:
         pass
 
     try:
         pdfobj = Array(obj)
         if pdfobj.is_rectangle:
-            return True
-    except Exception:
-        pass
+            return
+    except Exception as e:
+        raise ValueError("object is not a rectangle") from e
 
     raise ValueError("object is not a rectangle")
 
@@ -995,6 +972,7 @@ def check_is_box(obj) -> bool:
 class Extend_Page:
     @property
     def mediabox(self):
+        "This page's /MediaBox, in PDF units."
         return self._get_mediabox(True)
 
     @mediabox.setter
@@ -1004,6 +982,10 @@ class Extend_Page:
 
     @property
     def cropbox(self):
+        """This page's effective /CropBox, in PDF units.
+
+        If the /CropBox is not defined, the /MediaBox is returned.
+        """
         return self._get_cropbox(True)
 
     @cropbox.setter
@@ -1013,6 +995,11 @@ class Extend_Page:
 
     @property
     def trimbox(self):
+        """This page's effective /TrimBox, in PDF units.
+
+        If the /TrimBox is not defined, the /CropBox is returned (and if
+        /CropBox is not defined, /MediaBox is returned).
+        """
         return self._get_trimbox(True)
 
     @trimbox.setter
@@ -1021,7 +1008,13 @@ class Extend_Page:
         self.obj['/TrimBox'] = value
 
     @property
-    def resources(self):
+    def images(self) -> _ObjectMapping:
+        """Return all images associated with this page."""
+        return self._images
+
+    @property
+    def resources(self) -> Dictionary:
+        """Return this page's resources dictionary."""
         return self.obj['/Resources']
 
     def add_resource(
@@ -1032,14 +1025,15 @@ class Extend_Page:
         *,
         prefix: str = '',
         replace_existing: bool = True,
-    ) -> str:
+    ) -> Name:
         """Adds a new resource to the page's Resources dictionary.
 
         If the Resources dictionaries do not exist, they will be created.
 
         Args:
             self: The object to add to the resources dictionary.
-            res: The resource dictionary object to add.
+            res: The dictionary object to insert into the resources
+                dictionary.
             res_type: Should be one of the following Resource dictionary types:
                 ExtGState, ColorSpace, Pattern, Shading, XObject, Font, Properties.
             name: The name of the object. If omitted, a random name will be
@@ -1053,17 +1047,22 @@ class Extend_Page:
             The name of the object.
 
         Example:
-            >>> resource_name = Page(pdf.pages[0]).add_resource(formxobj, Name.XObject)
+            >>> resource_name = pdf.pages[0].add_resource(formxobj, Name.XObject)
 
-        .. versionadded: 2.3
+        .. versionadded:: 2.3
+
+        .. versionchanged:: 2.14
+            If *res* does not belong to the same `Pdf` that owns this page,
+            a copy of *res* is automatically created and added instead. In previous
+            versions, it was necessary to change for this case manually.
         """
-        if not Name.Resources in self.obj:
+        if Name.Resources not in self.obj:
             self.obj.Resources = Dictionary()
         elif not isinstance(self.obj.Resources, Dictionary):
             raise TypeError("Page /Resources exists but is not a dictionary")
         resources = self.obj.Resources
 
-        if not res_type in resources:
+        if res_type not in resources:
             resources[res_type] = Dictionary()
 
         if name is not None and prefix:
@@ -1080,11 +1079,127 @@ class Extend_Page:
                 else:
                     raise ValueError(f"Name {name} already exists in page /Resources")
 
-        resources[res_type][name] = res
+        resources[res_type][name] = res.with_same_owner_as(self.obj)
         return name
 
+    def _over_underlay(
+        self, other, rect: Optional[Rectangle], under: bool = True
+    ) -> None:
+        formx = None
+        if isinstance(other, Page):
+            page = other
+            formx = other.as_form_xobject()
+        elif isinstance(other, Dictionary) and other.get(Name.Type) == Name.Page:
+            page = Page(other)
+            formx = page.as_form_xobject()
+        elif (
+            isinstance(other, Stream)
+            and other.get(Name.Type) == Name.XObject
+            and other.get(Name.Subtype) == Name.Form
+        ):
+            formx = other
+
+        if formx is None:
+            raise TypeError("other object is not something we can convert to FormX")
+
+        if rect is None:
+            rect = Rectangle(page.trimbox)
+
+        formx_placed_name = self.add_resource(formx, Name.XObject)
+        cs = self.calc_form_xobject_placement(formx, formx_placed_name, rect)
+
+        self.contents_add(cs, prepend=under)
+
+    def add_overlay(self, other: Union[Object, Page], rect: Optional[Rectangle] = None):
+        """Overlay another object on this page.
+
+        Overlays will be drawn after all previous content, potentially drawing on top
+        of existing content.
+
+        Args:
+            other: A Page or Form XObject to render as an overlay on top of this
+                page.
+            rect: The PDF rectangle (in PDF units) in which to draw the overlay.
+                If omitted, this page's trimbox, cropbox or mediabox will be used.
+
+        .. versionadded:: 2.14
+        """
+        return self._over_underlay(other, rect, under=False)
+
+    def add_underlay(
+        self, other: Union[Object, Page], rect: Optional[Rectangle] = None
+    ):
+        """Underlay another object beneath this page.
+
+        Underlays will be drawn before all other content, so they may be overdrawn
+        partially or completely.
+
+        Args:
+            other: A Page or Form XObject to render as an underlay underneath this
+                page.
+            rect: The PDF rectangle (in PDF units) in which to draw the underlay.
+                If omitted, this page's MediaBox will be used.
+
+        .. versionadded:: 2.14
+        """
+        return self._over_underlay(other, rect, under=True)
+
+    def contents_add(self, contents: Union[Stream, bytes], *, prepend: bool = False):
+        """Append or prepend to an existing page's content stream.
+
+        Args:
+            contents: An existing content stream to append or prepend.
+            prepend: Prepend if true, append if false (default).
+
+        .. versionadded:: 2.14
+        """
+        return self._contents_add(contents, prepend=prepend)
+
+    def __getattr__(self, name):
+        return getattr(self.obj, name)
+
+    @augment_override_cpp
+    def __setattr__(self, name, value):
+        if hasattr(self.__class__, name):
+            return object.__setattr__(self, name, value)
+        setattr(self.obj, name, value)
+
+    @augment_override_cpp
+    def __delattr__(self, name):
+        if hasattr(self.__class__, name):
+            return object.__delattr__(self, name)
+        delattr(self.obj, name)
+
+    def __getitem__(self, key):
+        return self.obj[key]
+
+    def __setitem__(self, key, value):
+        self.obj[key] = value
+
+    def __delitem__(self, key):
+        del self.obj[key]
+
+    def __contains__(self, key):
+        return key in self.obj
+
+    def __eq__(self, other):
+        return self.obj == other.obj
+
+    def get(self, key, default=None):
+        try:
+            return self[key]
+        except KeyError:
+            return default
+
+    def emplace(self, other: Page, retain=(Name.Parent,)):
+        return self.obj.emplace(other.obj, retain=retain)
+
     def __repr__(self):
-        return repr(self.obj).replace('Dictionary', 'Page', 1)
+        return (
+            repr(self.obj)
+            .replace('Dictionary', 'Page', 1)
+            .replace('(Type="/Page")', '', 1)
+        )
 
     def _repr_mimebundle_(self, include=None, exclude=None):
         data = {}
@@ -1108,3 +1223,159 @@ class Extend_Page:
 class Extend_Token:
     def __repr__(self):
         return f'pikepdf.Token({self.type_}, {self.raw_value})'
+
+
+@augments(Rectangle)
+class Extend_Rectangle:
+    def __repr__(self):
+        return f'pikepdf.Rectangle({self.llx}, {self.lly}, {self.urx}, {self.ury})'
+
+    def __hash__(self):
+        return hash((self.llx, self.lly, self.urx, self.ury))
+
+
+@augments(Attachments)
+class Extend_Attachments(MutableMapping):
+    def __getitem__(self, k: str) -> AttachedFileSpec:
+        filespec = self._get_filespec(k)
+        if filespec is None:
+            raise KeyError(k)
+        return filespec
+
+    def __setitem__(self, k: str, v: AttachedFileSpec) -> None:
+        if not v.filename:
+            v.filename = k
+        return self._add_replace_filespec(k, v)
+
+    def __delitem__(self, k: str) -> None:
+        return self._remove_filespec(k)
+
+    def __len__(self):
+        return len(self._get_all_filespecs())
+
+    def __iter__(self) -> Iterator[str]:
+        yield from self._get_all_filespecs()
+
+    def __repr__(self):
+        return f"<pikepdf._qpdf.Attachments with {len(self)} attached files>"
+
+
+@augments(AttachedFileSpec)
+class Extend_AttachedFileSpec:
+    @staticmethod
+    def from_filepath(pdf: Pdf, path: Union[Path, str], *, description: str = ''):
+        """Construct a file specification from a file path.
+
+        This function will automatically add a creation and modified date
+        using the file system, and a MIME type inferred from the file's extension.
+
+        Args:
+            pdf: The Pdf to attach this file specification to.
+            path: A file path for the file to attach to this Pdf.
+            description: An optional description. May be shown to the user in
+                PDF viewers.
+        """
+        mime, _ = mimetypes.guess_type(str(path))
+        if mime is None:
+            mime = ''
+        if not isinstance(path, Path):
+            path = Path(path)
+
+        stat = path.stat()
+        return AttachedFileSpec(
+            pdf,
+            path.read_bytes(),
+            description=description,
+            filename=str(path),
+            mime_type=mime,
+            creation_date=encode_pdf_date(
+                datetime.datetime.fromtimestamp(stat.st_ctime)
+            ),
+            mod_date=encode_pdf_date(datetime.datetime.fromtimestamp(stat.st_mtime)),
+        )
+
+    def __repr__(self):
+        if self.filename:
+            return (
+                f"<pikepdf._qpdf.AttachedFileSpec for {self.filename!r}, "
+                f"description {self.description!r}>"
+            )
+        else:
+            return f"<pikepdf._qpdf.AttachedFileSpec description {self.description!r}>"
+
+
+@augments(AttachedFile)
+class Extend_AttachedFile:
+    @property
+    def creation_date(self) -> Optional[datetime.datetime]:
+        if not self._creation_date:
+            return None
+        return decode_pdf_date(self._creation_date)
+
+    @creation_date.setter
+    def creation_date(self, value: datetime.datetime):
+        self._creation_date = encode_pdf_date(value)
+
+    @property
+    def mod_date(self) -> Optional[datetime.datetime]:
+        if not self._mod_date:
+            return None
+        return decode_pdf_date(self._mod_date)
+
+    @mod_date.setter
+    def mod_date(self, value: datetime.datetime):
+        self._mod_date = encode_pdf_date(value)
+
+    def read_bytes(self) -> bytes:
+        return self.obj.read_bytes()
+
+    def __repr__(self):
+        return (
+            f'<pikepdf._qpdf.AttachedFile objid={self.obj.objgen} size={self.size} '
+            f'mime_type={self.mime_type} creation_date={self.creation_date} '
+            f'mod_date={self.mod_date}>'
+        )
+
+
+@augments(NameTree)
+class Extend_NameTree(MutableMapping):
+    def __len__(self):
+        return len(self._as_map())
+
+    def __iter__(self):
+        for name, _value in self._nameval_iter():
+            yield name
+
+    def keys(self):
+        return KeysView(self._as_map())
+
+    def values(self):
+        return ValuesView(self._as_map())
+
+    def items(self):
+        return ItemsView(self._as_map())
+
+    def __eq__(self, other):
+        return self.obj.objgen == other.obj.objgen
+
+    def __contains__(self, name: Union[str, bytes]) -> bool:
+        """
+        Returns True if the name tree contains the specified name.
+
+        Args:
+            name (str or bytes): The name to search for in the name tree.
+                This is not a PDF /Name object, but an arbitrary key.
+                If name is a *str*, we search the name tree for the UTF-8
+                encoded form of name. If *bytes*, we search for a key
+                equal to those bytes.
+        """
+        return self._contains(name)
+
+    def __getitem__(self, name: Union[str, bytes]) -> Object:
+        return self._getitem(name)
+
+    def __setitem__(self, name: Union[str, bytes], o: Object):
+        self._setitem(name, o)
+
+    def __delitem__(self, name: Union[str, bytes]):
+        self._delitem(name)
